/*****************************************************************************/

/*
 *      dabcore.c  --  dab pci driver.
 *
 *      Copyright (C) 1999-2005  Deti Fliegl, Georg Acher ({fliegl|acher}@baycom.de)
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *  Changes compared to 2.4-dabcore
 *  - No longer OSS dependant, no mixer ioctls 
 *  - Analogue audio volume set to default value
 *  
 *
 *  ATTENTION: The Thesys chip does not support shared interrupts!
 *
 *
 *  $Id: dabcore.c,v 1.1.2.1 2006/05/11 11:22:44 acher Exp $
 *
 */

/*****************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/socket.h>
#include <linux/list.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <asm/uaccess.h>
#include <asm/atomic.h>
#include <linux/delay.h>
#include <linux/smp_lock.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/cdev.h>
#include <linux/device.h>

#include "dabapi.h"
#include "dabcore.h"
#define VERSTR "version v2.01 time " __TIME__ " " __DATE__ 

//#define DEBUG

#define DABCORE_MAJOR 242 
#define MAX_DABCORE_MINORS 4

#define MODNAME "dabcore"
#define MODSTR MODNAME": "

#define MB mb();

#ifdef DEBUG
#define dbg(format, arg...)  printk(KERN_DEBUG MODSTR format"\n", ## arg)
#else
#define dbg(format, arg...)  
#endif

#define info(format, arg...) printk(KERN_INFO MODSTR format"\n", ## arg)
#define err(format, arg...) printk(KERN_ERR MODSTR format"\n", ## arg)

#define PCI_DABCORE_SIZE 4096
#define DMABUF_DEFAULTORDER 16
#define DMABUF_MINORDER 4
#define BESTAT_READ 100

#define STOP		0
#define STARTED		1
#define HALF		2
#define FULL		4

#define PCI_DEVICE_ID_THESYS_TH6255	0x186f 
#define PCI_VENDOR_ID_THESYS		0x10fb
#define NR_DEVICE 4

#define TH6255_EXTEND 0x34

// ------------ Control & Status Registers 

#define BECON		0x0		// 0x0 Back End Control Register
#define BESTAT		0x4		// 0x4 Back End Status Register
#define ACONREG		0x8		// 0x8 Audio Control Register
#define VCONREG 	0xc		// 0xC Video Control Register
#define AMEMBP  	0x10		// 0x10 Audio Memory Base Pointer
#define VMEMBP1		0x14		// 0x14 Video Memory Base Pointer 1
#define VMEMBP2		0x18		// 0x18 Video Memory Base Pointer 2
#define VMEMSTR		0x1c		// 0x1c Video Line Offset
#define VMSKBP1		0x20		// 0x20 Video Mask Memory Base Pointer 1
#define VMSKBP2		0x24		// 0x24 Video Mask Memory Base Pointer 2
#define VMSKSTR		0x28		// 0x28 Video Line Offset for Mask Map
#define I2CIMREG	0x2c		// 0x2c I2C / IM Interface Register
#define GPIOREG		0x30		// 0x30 General Purpose Port Register

// ------------ BECON-Bits

#define VCLPOL		0x20000		// Video Clock Polarity
#define SRESET		0x10000		// Software Reset
#define ARBITRATE	0x100		// Data Transfer Priority
#define IMSK7		0x80		// Mask Data Latency Error Interrupt
#define IMSK6		0x40		// Video Data Latency Error Interrupt
#define IMSK5		0x20		// Audio Memory Plane 1 End Interrupt
#define IMSK4		0x10		// Audio Memory Plane 0 End Interrupt
#define IMSK3		0x8		// GP-I/O Interrupt
#define IMSK2		0x4		// Audio FIFO-Full Interrupt
#define IMSK1		0x2		// Video FIFO-Full Interrupt
#define IMSK0		0x1		// Video Vertical Syncro INterrupt

// ------------ BESTAT-Bits

#define STOVI		0x1000000	// Video Stream Stopped
#define VFRAME		0x100000	// Video Field ID
#define RECFORM		0x40000		// 2Bit Audioformat
#define ASHP		0x100		// 8Bit Audio Word Count bit 17...8
#define INTMDL		0x80		// Mask Data Latency Error
#define INTVDL		0x40		// Video Latency Error
#define INTAMPE0	0x20		// Audio Memory Plane 0 End Interrupt
#define INTAMPE1	0x10		// Audio Memory Plane 1 End Interrupt
#define INTGP		0x8		// GP/IO Interrupt
#define INTAFF		0x4		// Audio FIFO-Full Interrupt
#define INTVFF		0x2		// Video FIFO-Full Interrupt
#define INTVIS		0x1		// Video Vertical Sync

// ------------ ACONREG-Bits

#define APSIZE		0x8000		// 16Bit Audiobuffersize
#define ADWMIN		0x100		// 5Bit min. Dwords in Fifo to start pci burst
#define WSLL		0x20		// I2S WordStrobe Left Indicator
#define SENDFORM	0x8		// 2Bits Playback Audioformat
#define SONYPHIMO	0x4		// Sony / Philis I2S mode
#define PLAYRECMO	0x2		// AudioMode Rec/Play
#define STARTAD		0x1		// AudioDataEnable

// ------------ VCONREG-Bits

#define DWNUM		0x100		// 11Bit min. Dwords in Fifo to start pci burst
#define FHPIC		0x80 		// Video Mode Set
#define BYORD		0x40		// Byte Order Set
#define UNPAC		0x20		// RGB888 packed/unpacked
#define VALFA		0x10		// alpha-key enable
#define PIXFORM		0x4		// 2Bit Pixelformat
#define STAMASK		0x2		// Video Mask On/Off
#define STARTVD		0x1		// Video On/Off

// ------------ I2CIMREG-Bits

#define SCL		0x4		// SCL Line control
#define SDA		0x2		// SDA Line control
#define IMIDENT		0x1		// IM Bus Address/Data Select

// ------------ GPIOREG-Bits

#define USERIDENT	0x1000000	// 8Bits Hardware Identity
#define GPIODAT		0x10000		// 8Bits General Purpose IO Dataport
#define GPOUT		0x100		// 4Bits General Purpose Output Bits
#define IOWEXN		0x1		// 8Bits Input/Output Select for GP/IO


// ------------ Lines to DAB-Module

#define MCE		GPOUT		// TST0 Module Chip Enable
#define MOE		(GPOUT<<1)	// TST1 Module Output Enable
#define LATCHLOW	(GPOUT<<2)	// TST2 Latch for MA0-MA7
#define LATCHHIGH	(GPOUT<<3)	// TST3 Latch for MA8-MA9, SRES, CSDA
#define SRES		(GPIODAT<<6)	// GPIO6 SRES Line to DAC080 or CS84*
#define CSDA		(GPIODAT<<7)	// GPIO7 CSDA Line to DAC080
#define MRW		IMIDENT		// IMIDENT Module R/W control
#define SCLK		SCL		// SCL Clockline to DAC080
#define CDIN		SDA		// SDA Dataline to DAC080
#define MINT		INTGP		// INTGP Interrupt line from Module


// ------------ DAB-Module registers

#define MSGREG_DAB	0x3ff		// Message from DAB Module to Controller
#define MSGREG_CPU	0x3fe		// Message from Controller to DAB Module
#define DATA_TYPE	0x3fd		// Data Type of BUFFPTR_FIC 0:empty 1:FIC 2:PAD 
#define FREEPTR_DAB	0x3fb		// First free field in circular buffer DAB Module -> Controller
#define DATAPTR_DAB	0x3fa		// Lowest used field in circular buffer DAB Module -> Controller
#define FREEPTR_CPU	0x3f9		// First free field in circular buffer Controller -> DAB Module
#define DATAPTR_CPU	0x3f8		// Lowest used field in circular buffer -> DAB Module
#define BUFFPTR_CPU	960		// circular buffer Controller -> DAB Module
#define BUFSIZE_CPU	32		// size in bytes
#define BUFFPTR_DAB	768		// circular buffer DAB Module -> Controller
#define BUFSIZE_DAB	192		// size in bytes
#define BUFFPTR_FIC	0		// FIC/PAD Databuffer
#define BUFSIZE_FIC	768		// size in bytes

/* --------------------------------------------------------------------- */

#define BUFSIZE 8192

static int total_bufsize=256*1024; //256kbyte
static int buffers=32; //32*8k = 256Kbytes

/* --------------------------------------------------------------------- */

typedef struct buff buff_t;
struct buff
{
	buff_t *next;
	void *ptr;
	unsigned int state;
};
typedef buff_t *  pbuff_t;

typedef struct buffq
{	
	pbuff_t q;
	pbuff_t end;
	wait_queue_head_t wait;
} buffq_t;

#define	empty(x) while(add_buf_end(s, &s->free,x));	// empty queue 

static int emuirq=0;
static int pciburstlen=1;
typedef struct dabcore dabcore_t;
struct dabcore
{
	/* hardware resources */
	unsigned int irq;
	void *mem;
	unsigned int *audiobuf;
	int audiobufsize;
	struct pci_dev *dev;
	
	/* buffers 	*/	
	buffq_t free;
	buffq_t rec;
	
	unsigned int readptr;
	/* state */
	int state;
	int irqstate;
	int gotmem;
	int sres_csda;
	int irqcount;
	int newmsg;
	wait_queue_head_t msgwait;
	int mix_vol;
	int mix_modcnt;
	int aconreg;
	pbuff_t intbuf;
	int overflow;
	spinlock_t lock;
	spinlock_t lockbuf;
	struct semaphore open_sem;
	mode_t open_mode;
	wait_queue_head_t open_wait;
	dma_addr_t rx_bufs_dma;
	void *rx_bufs;
}; 
/* --------------------------------------------------------------------- */

static dabcore_t *dabcore_devs[MAX_DABCORE_MINORS];
static int dabcore_count=0;
static struct class *dabcore_class;

static struct timer_list irq_emu_timer;

/* --------------------------------------------------------------------- */

#ifdef DEBUG
void beep(long freq)
{
  long v;
  char low,high;

  if(!freq)
    outb( inb( 0x61 )&252, 0x61 );
  else
  {
    outb( inb( 0x61 ) | 0x3, 0x61 );

    v=1193180L/freq;

    low=(char)(v&255);
    high=(char)((v>>8)&255);

    outb(182, 0x43);
    outb(low, 0x42);
    outb(high, 0x42);
  }
}

void showbufs(dabcore_t *s)
{
	pbuff_t this=s->free.q;
	printk(KERN_DEBUG MODSTR"------------------------\n");	
	printk(KERN_DEBUG MODSTR"Buffer statistics\n");
	printk(KERN_DEBUG MODSTR"free buffers: free_end %p\n", s->free.end);
	while(this)				 
	{
		printk(KERN_DEBUG MODSTR"free buffer %08x ptr %p\n",(unsigned int)this, this->ptr);
		this=this->next;
	}
	this=s->rec.q;
	printk(KERN_DEBUG MODSTR"------------------------\n");		
	printk(KERN_DEBUG MODSTR"rec buffers: rec_end %p\n",s->rec.end);
	while(this)				 
	{
		printk(KERN_DEBUG MODSTR"rec buffer %08x ptr %p\n",(unsigned int)this, this->ptr);
		this=this->next;
	}
	printk(KERN_DEBUG MODSTR"------------------------\n");	
}
#endif

/* --------------------------------------------------------------------- */

static int add_buf_end(dabcore_t *s, buffq_t *dst, buffq_t *src)
{
	pbuff_t tmp;
	unsigned long flags;


	dbg("add_buf_end: %p %p",dst->q, src->q);

	if(src->q==NULL)		// no more buffers available
		return 0;

	spin_lock_irqsave(&s->lockbuf, flags);
	tmp=src->q;
	src->q=src->q->next;		// remove buffer from src
	
	if(src->q==NULL)		// end of source reached
		src->end=NULL;		// update end pointer.

	tmp->state=0;			// delete any state
	tmp->next=NULL;			// src will be added at end of dstchain so nextptr is NULL

	if(dst->end==NULL)
		dst->q=tmp;		// first buffer in queue
	else
		dst->end->next=tmp;	// add new buffer at end of queue

	dst->end=tmp;			// set end to new position

	spin_unlock_irqrestore(&s->lockbuf, flags);	
	return 1;
}

/* --------------------------------------------------------------------- */

static void wait_eoi(dabcore_t *s)
{
	if(s->rec.end)
	{
		unsigned long x=jiffies;
		unsigned long end=(BUFSIZE*HZ)/384000;
		while((jiffies-x)<=end)
			schedule();
	}
}

/* --------------------------------------------------------------------- */


static int free_buffers(dabcore_t * s)
{
	buff_t tmp;
	//dbg("free_buffers\n");
	if(!s->gotmem)
		return 0;
		
	empty(&s->rec);
	while(s->free.q)				// free freequeue 
	{
		tmp=*(s->free.q);			// copy first element to tmp
		kfree(s->free.q);			// free struct
		s->free.q=tmp.next;			// next element to first element
	}
	s->gotmem=0;
	if (s->rx_bufs){
			pci_free_consistent(s->dev, total_bufsize , s->rx_bufs, s->rx_bufs_dma);
        }
	return 1;
}

static int alloc_buffers(dabcore_t *s)
{
	pbuff_t new;
	int first=1;
	int i;
	
       	dbg("alloc_buffers %x %x\n",gotmem,total_bufsize);

	s->rx_bufs = pci_alloc_consistent(s->dev, total_bufsize, &s->rx_bufs_dma);
       	
       	if (s->rx_bufs == NULL) {
       	  return 0;
       	}
       	
	for( i=0 ; i<buffers ; i++)
	{
        	dbg("xalloc_buffers %x %x\n",gotmem, total_bufsize);
		new=kmalloc(sizeof(buff_t), GFP_KERNEL);// Alloc struct
		if(new==NULL) {
		        printk("failed to alloc buff_t struct\n");
			break;
                }
			
		new->next=NULL;				// init
		new->state=0;

		if(first)
		{
			s->free.end=new;		// set end
			first=0;
		}
		
		new->ptr=s->rx_bufs+i*BUFSIZE;
		new->next=s->free.q;			// insert as first element
		s->free.q=new;
	}
	if(i < buffers)
	{
		free_buffers(s);			// clean up if failure
		return 0;
	}
	s->gotmem=1;
	#ifdef DEBUG
	showbufs(s);
	#endif
	return 1;
}


/* --------------------------------------------------------------------- */
#define writereg(s, reg, value)	{MB;writel(value, (volatile int*)(s->mem+(reg)));MB;}

inline int readgpio(dabcore_t *s)
{
	int i;
	int ret=0;
	unsigned long flags;
	MB;
	spin_lock_irqsave(&s->lock, flags);
	for(i=0; i<5; i++) 
	{
	  ret=readl((volatile int*)(s->mem+GPIOREG));
	  //printk(MODSTR "try: %2d ret: %08X\n",i, ret);
		if((i>1)&&(ret&0xff0000ff)==0xff0000ff)
			break;
	}
	spin_unlock_irqrestore(&s->lock, flags);
	return ret;
}
/* --------------------------------------------------------------------- */

static void address2latch(dabcore_t *s, int addr)
{
	int latchlow=addr&0xff;			// store lowbyte of address
	int latchhigh=(addr>>8)&0x3;		// store highbyte of address

	//dbg("address2latch\n");
	writereg(s, GPIOREG, (latchlow<<16) | MOE | MCE);
	writereg(s, GPIOREG, (latchlow<<16) | MOE | MCE | LATCHLOW );	// write lowbyte to lowlatch
	writereg(s, GPIOREG, (latchlow<<16) | MOE | MCE);
	writereg(s, GPIOREG, (latchhigh<<16) | s->sres_csda | MOE | MCE);
	writereg(s, GPIOREG, (latchhigh<<16) | s->sres_csda | MOE | MCE | LATCHHIGH);	// write lowbyte to lowlatch
	writereg(s, GPIOREG, (latchhigh<<16) | s->sres_csda | MOE | MCE);
	writereg(s, GPIOREG, MOE | MCE);	// set GPIO to output
}

static void write_module(dabcore_t *s, int addr, int val)
{
	//dbg("writemodule\n");
	val&=0xff;
	address2latch(s,addr);
	writereg(s, I2CIMREG, SCLK|SDA);		// Generate write cycle
	writereg(s, GPIOREG, (val<<16) | MOE | MCE);	// output data 
	writereg(s, GPIOREG, (val<<16) | MOE);	// output data and assert MCE
	writereg(s, GPIOREG, MOE | MCE);	// deassert MCE.
	writereg(s, I2CIMREG, MRW | SCLK|SDA);	// Cancel write cycle	
}

static int read_module(dabcore_t *s, int addr)
{
	int ret;
	
	//dbg("readmodule\n");	

	address2latch(s,addr);
	writereg(s, I2CIMREG, MRW | SCLK | CDIN|SDA);	// Generate read cycle	
	writereg(s, GPIOREG, 0xff | MOE | MCE);	// input data	
	writereg(s, GPIOREG, 0xff);		// input data and assert MCE and MOE
	ret=(readgpio(s)>>16) & 0xff;	// Read GPIODATA lines
	writereg(s, GPIOREG, MOE | MCE);	// deassert MCE.

	return ret;
}
/* --------------------------------------------------------------------- */

static void dac_serialize(dabcore_t *s, int val)
{
	int i;
	int reg;
	
	//dbg("dac_serialize\n");
	for(i=0;i<8;i++)		// send addr msb first
	{
		if(val&0x80)
			reg=CDIN;
		else
			reg=0;
		writereg(s, I2CIMREG, MRW | reg);
		udelay(1);
		writereg(s, I2CIMREG, MRW | SCLK | reg);
		udelay(1);
		val<<=1;
	}
		writereg(s, I2CIMREG, MRW | SCLK | SDA);
}

static void write_dacctrl(dabcore_t *s, int addr, int val)
{
	//dbg("write_dacctrl\n");
	writereg(s, I2CIMREG, MRW | SCLK);	// set SCLK to high
	s->sres_csda&=~CSDA;		// assert chipselect for DAC
	address2latch(s,0);		// generate an address cycle to set latches
	dac_serialize(s,addr|0x80);	// first serialize controlword
	dac_serialize(s,val);		// then databyte
	s->sres_csda|=CSDA;		// de-asseert chipselect for DAC
	address2latch(s,0);		// generate an address cycle to set latches
	writereg(s, I2CIMREG, MRW | SCLK | SDA);
}

static void dacrate(dabcore_t *s, int val)
{
	if(val)
		// Source Selection Control Register
		write_dacctrl(s, 1, 0x89);		// Main serial PCM Data, MCLK2
	else
		write_dacctrl(s, 1, 0x81);		// Main serial PCM Data, MCLK1
}

static int dacvol(dabcore_t *s, int vol)
{
	int leftvol=(vol&0xff)*69/100;
	int rightvol=((vol>>8)&0xff)*69/100;

	//printk(KERN_DEBUG MODSTR"vol %d %d %d\n",vol, leftvol, rightvol);
	// Volume Control
	write_dacctrl(s, 3, leftvol|0x80);		
	write_dacctrl(s, 4, rightvol);
	s->mix_vol=vol;
	return 0;
}

/* --------------------------------------------------------------------- */

int read_module_buf(dabcore_t *s, unsigned char *buf)
{
	int start=read_module(s, DATAPTR_DAB);
	int end=read_module(s, FREEPTR_DAB);
	int cnt=0;
	
	//printk(KERN_DEBUG MODSTR"read_module_buf %d %d\n", start, end);
	
	read_module(s, MSGREG_DAB);		// be sure we gained access

	if((end%BUFSIZE_DAB)!=end)
	{
		dbg("read_module_buf Communication error.");
		return 0;
	}
		
	while(start!=end)
	{
		*buf=read_module(s, start+BUFFPTR_DAB);
		start=(start+1)%BUFSIZE_DAB;
		buf++;
		cnt++;
	}
	
	//write_module(s, DATAPTR_DAB, end);
	
	write_module(s, DATAPTR_DAB, 0);
	write_module(s, FREEPTR_DAB, 0);
	
	//printk("writing %d %x\n",DATAPTR_DAB,end);
	return cnt;
}

int write_module_buf(dabcore_t *s, unsigned char *buf, int size)
{
	int start=read_module(s, FREEPTR_CPU);
	int end=read_module(s, DATAPTR_CPU);
	int cnt=start;
	int ret=0;
	
	if(size > BUFSIZE_CPU)
		size=BUFSIZE_CPU;

	read_module(s, MSGREG_DAB);		// be sure we gained access

	while(size--)
	{
		write_module(s, cnt+BUFFPTR_CPU, *buf);
		cnt=(cnt+1)%BUFSIZE_CPU;
		buf++;
		ret++;
		if(cnt==end)
			break;
	}
	write_module(s, FREEPTR_CPU, cnt);
	//write_module(s, DATAPTR_CPU, start);
	//s->newmsg=0;

	//read_module(s, MSGREG_DAB);
	//write_module(s, MSGREG_DAB, 0x0);	
	write_module(s, MSGREG_CPU, 0xf);
	return ret;
}

/* --------------------------------------------------------------------- */

static void showregs(dabcore_t *s)
{
	int i;
	
	printk(KERN_DEBUG MODSTR"PCI Bridge Registers:\n");
	for(i=0;i<13;i++)
		printk(KERN_DEBUG MODSTR"Reg: 0x%02x Val: %#x\n", i<<2, readl((volatile int*)(s->mem+(i<<2))));
}

static void show_module_regs(dabcore_t *s)
{
	int i;
	//printk(KERN_DEBUG MODSTR"DAB Module DP-RAM:");
	for(i=0; i<1024;i++)
	{
		if(!(i&0xf))
			printk("\n" KERN_DEBUG MODSTR "Module %04d:", i);
		printk(" %02X",read_module(s,i));
	}
	printk("\n");
}

/* --------------------------------------------------------------------- */

static void init_dac(dabcore_t *s)
{	
	dbg("init_dac\n");
	// Control Register
	write_dacctrl(s, 2, 0xc1);		// no deemphasis, device disabled
	mdelay(5);
	write_dacctrl(s, 2, 0xc0);		// no deemphasis, device enabled
	// Serial Data Format Control Register
	write_dacctrl(s, 0, 0x3);		// I2S Philips, 16bit, MCLK=256xLRCLK, BITCLK=32xLRCLK
	// Source Selection Control Register
	write_dacctrl(s, 1, 0x89);		// Main serial PCM Data, MCLK2
	// Volume Control
	dacvol(s, 0x4242);			// set to Vol left + right to +6dB
}

static void init_dab_module(dabcore_t *s)
{
	dbg("init_dab_module\n");
	// GPIO Lines
	writereg(s, GPIOREG, MCE | MOE);	// set MCE MOE to high and GPIODAT to output
	s->sres_csda=CSDA;			// set CSDA to high reset DAB Module
	address2latch(s, 0);			// generate an address cycle to set latches
	mdelay(1);
	s->sres_csda=SRES | CSDA;		// set SRES CSDA to high
	address2latch(s, 0);			// generate an address cycle to set latches
	writereg(s, I2CIMREG, MRW | SCLK);		
	writereg(s, I2CIMREG, MRW | SCLK|SDA);		
}

static void init_chipset(dabcore_t *s)
{
	int i;
	
	dbg("init_chipset\n");
	
	writereg(s, BECON, SRESET);	//initalize PCI-Bridge
	udelay(100);
	writereg(s, BECON, 0);		// return from software reset
	
	// now switch to standard I2S with 24Bit REC-mode
	// aconreg is buffered in the context structure to avoid reading
	// the buggy PCI-bridge
	s->aconreg=(pciburstlen<<8) | PLAYRECMO | SONYPHIMO;
	writereg(s, ACONREG, s->aconreg);
	writereg(s, BECON, IMSK3  | IMSK4 | IMSK5 ); // Enable Interrupt for GPIO, Audio Plane0 and Plane 1
#if DEBUG
	showregs(s);
#endif
	writereg(s, I2CIMREG, MRW | SCLK|SDA);
	init_dab_module(s);	
	init_dac(s);
	
	read_module(s, MSGREG_DAB);	// be sure we gained access 
	
	/*
	for(i=0; i<0x3fe ; i++)
		write_module(s, i, 0);
	*/	
	for(i=1016; i<1020 ; i++)
		write_module(s, i, 0);
	
}

static int intcnt=0;

/* --------------------------------------------------------------------- */
void stop(dabcore_t *s)
{
	s->aconreg&=~STARTAD;
	writereg(s, ACONREG, s->aconreg);
	s->state=STOP;
}

void startrek(dabcore_t *s)
{
	if(s->state)
		return;
	
	empty(&s->rec);
	if(add_buf_end(s, &s->rec, &s->free))
	{
		writereg(s, AMEMBP, virt_to_bus(s->rec.end->ptr));
		s->aconreg&=0x7fff;
		s->aconreg|=((((BUFSIZE>>2)-1)&0x1ffff)<<15)|STARTAD;
		writereg(s, ACONREG, s->aconreg);
		s->state=STARTED;
		s->readptr=0;
		intcnt=0;
	}
	else
		err(KERN_ERR MODSTR"dabcore: startrek: Paranoia (out of buffers).\n");
}

/* --------------------------------------------------------------------- */

static int dabcore_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
	dabcore_t *s = (dabcore_t *) dev_id;
	int i;
	int handled=0;

	writereg(s, BESTAT, 0);
	for(i=0; i<BESTAT_READ; i++) 
	{
		s->irqstate=readl(s->mem+BESTAT);
		if((s->irqstate&0xfff3ff00)==0x01000000)
			break;
	}

	if (i==BESTAT_READ) {
	        printk(MODSTR "i%d state %08x\n",i,s->irqstate);
		return IRQ_NONE;
	}
	
	writereg(s, BESTAT, s->irqstate);
	if(s->irqstate & INTGP)
	{
		s->newmsg++;
		wake_up(&s->msgwait);
		handled=1;
	}
	
	if(s->irqstate & INTAMPE0)
	{
		if(!s->rec.end)
		{
			err("Buffer overflow2");
			stop(s);
		}
		else
		{
			s->rec.end->state=HALF;
			s->intbuf=s->rec.end;
			if(add_buf_end(s, &s->rec, &s->free))
			{
				writereg(s, AMEMBP, virt_to_bus(s->rec.end->ptr));
				s->aconreg&=0x7fff;
				s->aconreg|=((((BUFSIZE>>2)-1)&0x1ffff)<<15)|STARTAD;
				writereg(s, ACONREG, s->aconreg);
				//intcnt++;
				//printk(MODSTR "dazu: %d\n",intcnt);
			}
			else
			{
				err("Buffer overflow1");
				s->overflow++;
				stop(s);
			}
		}
		wake_up(&s->rec.wait);		
		handled=1;
	}

	if(s->irqstate & INTAMPE1)
	{
		if(!s->rec.end)
		{
			stop(s);
			err("Buffer overflow3");
		}
		else
			s->intbuf->state=FULL;
		wake_up(&s->rec.wait);
		handled=1;
        }
//        if((s->irqstate&0xfff3ff00)!=0x01000000)
//      		  printk(MODSTR "Invalid interrupt state: %08X\n",s->irqstate);
		
	if (handled)
		return IRQ_HANDLED;
	else
		return IRQ_NONE;
}	
/* --------------------------------------------------------------------- */
void irq_emu_function(unsigned long dummy)
{
        int n;
        for(n=0;n<MAX_DABCORE_MINORS;n++) {
                if(dabcore_devs[n] && dabcore_devs[n]->dev)
                        dabcore_interrupt(0, dabcore_devs[n]->dev,NULL);
        } 
        mod_timer(&irq_emu_timer, jiffies+ (HZ/20));
}

/* --------------------------------------------------------------------- */

static ssize_t dabcore_read (struct file *file, char *buffer, size_t count, loff_t * ppos)
{
	dabcore_t *s = (dabcore_t *) file->private_data;
	ssize_t ret;
	unsigned cnt;
	unsigned rem;
	unsigned size=0;  //remove ugly compiler warning.
	int waitforsecondhalf=0;
	
	dbg("read\n");
	
/*	if (ppos != &file->f_pos)
		return -ESPIPE;
*/		
	if (!access_ok (VERIFY_WRITE, buffer, count))
		return -EFAULT;
	if(!s->gotmem && !alloc_buffers(s))
	{
		err("no space for buffers.");
		return -ENOBUFS;
	}

	ret = 0;
	//printk(MODSTR "start read\n");
	while (count > 0)
	{

	overflow_restart:
		startrek(s);
		if(s->rec.q && (s->rec.q->state & HALF) )		//paranoia check
			waitforsecondhalf=(s->readptr==(BUFSIZE>>1));	// read all data from first half?
		else
			waitforsecondhalf=0;
		
		if ( !s->rec.q || !s->rec.q->state || waitforsecondhalf)	// buffer is empty or wait for next one.
		{
			dbg("wait\n");
			//printk(MODSTR "go sleep\n");
			if (file->f_flags & O_NONBLOCK)		// return nonblocking
				return ret ? ret : -EAGAIN;
			interruptible_sleep_on(&s->rec.wait);	// otherwise sleep
			if (signal_pending(current))
				return ret ? ret : -ERESTARTSYS;
			if(!waitforsecondhalf)
				s->readptr=0;			// set readptr to 0
			//printk(MODSTR "woke up %d %d\n", s->state, count);
			if(s->state==STOP || !s->rec.q)
				goto overflow_restart;
			dbg("wait done\n");
		}
		if(!s->rec.q)
		{
			dbg("argl... s->rec.q==NULL");
			return 0;
		}
		
		if(!s->rec.q->ptr)
		{
			dbg("argl... s->rec.q->ptr==NULL");
			return 0;
		}
		size=(s->rec.q->state & HALF)?BUFSIZE>>1:BUFSIZE;

		rem=size-s->readptr;		// set remaining bytes to copy
		if(count >= rem)
			cnt=rem;
		else
			cnt=count;

		dbg(K"rec.q %p rec.q->ptr+s->readptr %p state %d wfsh %d readptr %d size %d cnt %d",
		    s->rec.q, s->rec.q->ptr+s->readptr, s->rec.q->state,waitforsecondhalf,s->readptr, size, cnt);

		if (copy_to_user(buffer, s->rec.q->ptr+s->readptr, cnt))
			return ret ? ret : -EFAULT;
			
		s->readptr+=cnt;
		count -= cnt;
		buffer += cnt;
		ret += cnt;
		//printk(MODSTR"readptr: %d rec.q->size: %d s->rec.q->state: %d size: %d cnt: %d\n",s->readptr,s->rec.q->size, s->rec.q->state,size, cnt);
		if(s->readptr==BUFSIZE)			// finished, take next buffer
		{
			dbg("new block\n");
			if(!add_buf_end(s, &s->free, &s->rec))
				dbg("read: add_buf_end =NULL");
			s->readptr=0;
			//intcnt--;
			//printk(MODSTR "weg: %d\n",intcnt);
		}
       	}
	return ret;
}

static int dabcore_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	void *ptr=(void *) arg;
	dabcore_t *s = (dabcore_t *) file->private_data;
	dab_dac_ctrl_t dab_dac_ctrl;
	dab_module_msg_t dab_module_msg;
	char buf[1024];
	int val;


	switch (cmd)
	{
		case IOCTL_DAB_DAC_CTRL:
			if (copy_from_user(&dab_dac_ctrl, ptr, sizeof(dab_dac_ctrl_t)))
				return -EFAULT;
			write_dacctrl(s, dab_dac_ctrl.addr, dab_dac_ctrl.val);
			return 0;
		
		case IOCTL_DAB_DAC_RATE:
			if (get_user(val, (int *) arg))
				return -EFAULT;
			dacrate(s, val);
			return 0;
			
		case IOCTL_DAB_WRITE_MSG:
			if (copy_from_user(&dab_module_msg , ptr, sizeof(dab_module_msg_t)))
				return -EFAULT;
			return write_module_buf(s, dab_module_msg.buf, dab_module_msg.size);
		
		case IOCTL_DAB_READ_MSG:
			memset(&dab_module_msg, 0, sizeof(dab_module_msg));
			dab_module_msg.size=read_module_buf(s, dab_module_msg.buf);
			return copy_to_user(ptr, &dab_module_msg, sizeof(dab_module_msg_t));

		case IOCTL_DAB_READ_MODULE:
			read_module(s, MSGREG_DAB);             // be sure we gained access
			for(val=0; val < 1024; val++)
				buf[val]=read_module(s, val);
			return copy_to_user(ptr, buf, sizeof(buf));

		case IOCTL_DAB_WRITE_MODULE:
			if (copy_from_user(buf, ptr, sizeof(buf)))
				return -EFAULT;
			for(val=0;val < 1024; val++)
				write_module(s, val, buf[val]);
			return 0;

		case IOCTL_DAB_READ_OK:
			if(!s->newmsg && !(file->f_flags & O_NONBLOCK) )
			{
				interruptible_sleep_on(&s->msgwait);
				if (signal_pending(current))
					return -ERESTARTSYS;
			}
			val=s->newmsg;
			s->newmsg=0;
			return put_user(val, (int *) arg);
		
		case IOCTL_DAB_OVERRUNS:
			return put_user(s->overflow, (int *) arg);

		case IOCTL_DAB_RDIFORMAT:
			val=(s->irqstate>>18)&3;
			return put_user(val, (int *) arg);			
		
		case IOCTL_DAB_RESETBUF:
			read_module(s, MSGREG_DAB);	// be sure we gained access 
			for(val=1016; val<1020 ; val++)
				write_module(s, val, 0);
			return 0;
				
		case IOCTL_DAB_INFO:
			return copy_to_user(ptr, VERSTR, sizeof(VERSTR)); 
		case IOCTL_DAB_VERSION:
			val=0;
			return put_user(val, (unsigned int *)arg);
			
		case IOCTL_DAB_WRITE_DPR:
			//get_user_ret (val, (int *) arg, -EFAULT);
			//write_module(s, val&255, val>>8);
			write_module(s, arg&1023, arg>>16);
			return 0;

		case IOCTL_DAB_READ_DPR:
			//get_user_ret (val, (int *) arg, -EFAULT);
			//val=read_module(s, val);
			//return put_user(val, (unsigned int *)arg);
			return read_module(s, arg);
		
		case IOCTL_DAB_DEBUG:
			
			/*if(!s->gotmem && !alloc_buffers(s, BUFFERNUM))
			{
				printk(KERN_ERR MODSTR"no space for buffers.\n");
				return -ENOBUFS;
			}
			startrek(s);
			*/
			return 0;
	}
	return -ENOIOCTLCMD;
}

static int dabcore_open (struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	dabcore_t *s = dabcore_devs[minor];
	
	if (!s)
		return -ENODEV;
		
	file->private_data = s;
	/* wait for device to become free */
	down (&s->open_sem);
	while (s->open_mode /*& file->f_mode*/)
	{
		if (file->f_flags & O_NONBLOCK)
		{
			up (&s->open_sem);
			return -EBUSY;
		}
		up (&s->open_sem);
		interruptible_sleep_on (&s->open_wait);
		if (signal_pending (current))
			return -ERESTARTSYS;
		down (&s->open_sem);
	}
	s->open_mode |= file->f_mode & (FMODE_READ | FMODE_WRITE);
	up (&s->open_sem);
	return 0;
}

static int dabcore_release (struct inode *inode, struct file *file)
{
	dabcore_t *s = (dabcore_t *) file->private_data;

	stop(s);
	down (&s->open_sem);
	s->open_mode &= (~file->f_mode) & (FMODE_READ | FMODE_WRITE);
	up (&s->open_sem);
	wake_up (&s->open_wait);
	return 0;
}

static struct file_operations dabcore_fops =
{
	owner: THIS_MODULE,
	read: &dabcore_read,
	ioctl: &dabcore_ioctl,
	open: &dabcore_open,
	release: &dabcore_release,
        llseek:  &no_llseek
};

/* --------------------------------------------------------------------- */

static int __devinit 
dabcore_pci_probe( struct pci_dev *pci_dev, const struct pci_device_id *pci_id )
{
	dabcore_t *s=NULL;
	long regstart;

	if (dabcore_count>=MAX_DABCORE_MINORS)
		return -ENODEV;

	if(pciburstlen<1 || pciburstlen>24)
	{
		err("Invalid pciburst value, setting to 1");
		pciburstlen=1;
	}

	total_bufsize=buffers*BUFSIZE;

	err("Using buffersize: %dkByte", total_bufsize>>10);

	s=(dabcore_t*)kmalloc(sizeof(dabcore_t),GFP_KERNEL);
	if (!s) {
		err("dabcore_pci_probe: Out of memory");
		goto probe_fail;
	}

	memset(s, 0, sizeof(dabcore_t));

	if ( pci_enable_device( pci_dev ) ) {
                err("dabcore_pci_probe: Can't enable PCI device");
                goto probe_fail;
        }

	regstart=pci_resource_start( pci_dev, 0 );

	if (!request_mem_region (regstart, PCI_DABCORE_SIZE, "dabcore")) {
                err("dabcore_pci_probe: MMIO resource %lx unavailable",regstart);
                goto probe_fail;
        }

        s->mem=ioremap(regstart, 16384);


	if (request_irq(pci_dev->irq, dabcore_interrupt, SA_SHIRQ,
			"dabcore", s)) {
		err("dabcore_pci_probe: Can't get IRQ %i",pci_dev->irq);
		iounmap(s->mem);
		release_mem_region(pci_resource_start(pci_dev, 0), PCI_DABCORE_SIZE);
		goto probe_fail;
	}
	
	s->irq=pci_dev->irq;
	s->dev=pci_dev;
	info("Bridge mapped at %p, IRQ %i",s->mem,s->irq);

	// Init structs

	init_waitqueue_head (&s->open_wait);
	init_waitqueue_head (&s->rec.wait);
	init_waitqueue_head (&s->free.wait);
	init_waitqueue_head (&s->msgwait);
	init_MUTEX(&s->open_sem);
	
	spin_lock_init(&s->lock);
	spin_lock_init(&s->lockbuf);
	
	init_chipset(s);

	dabcore_devs[dabcore_count]=s;
	dabcore_count++;

	pci_set_drvdata( pci_dev, s );

	return 0;

 probe_fail:
	if (s)
		kfree(s);
	return -ENODEV;
}

/* --------------------------------------------------------------------- */
static void __devexit dabcore_pci_remove( struct pci_dev *pci_dev )
{
	dabcore_t *s=(dabcore_t*)pci_get_drvdata( pci_dev );

	stop(s);
	writereg(s, BECON, SRESET);     //initalize PCI-Bridge
	udelay(100);
	writereg(s, BECON, 0);          // return from software reset
	wait_eoi(s);
	free_buffers(s);
	free_irq(pci_dev->irq, s);
	iounmap(s->mem);
	release_mem_region(pci_resource_start(pci_dev, 0), PCI_DABCORE_SIZE);
	free_buffers(s);
	pci_set_drvdata( pci_dev, NULL );
	kfree(s);
}

/* --------------------------------------------------------------------- */

static struct pci_device_id dabcore_pci_id[] = {
        {PCI_VENDOR_ID_THESYS, PCI_DEVICE_ID_THESYS_TH6255, PCI_ANY_ID, PCI_ANY_ID, },
        {0, }
};   

static struct pci_driver dabcore_driver =
{
        .name           = "dabcore",
        .id_table       = dabcore_pci_id,
        .probe          = dabcore_pci_probe,
        .remove         = __devexit_p(dabcore_pci_remove),
};



static int __init dabcore_init(void) 
{
	int i,err=0;

	if (register_chrdev (DABCORE_MAJOR, "dabcore", &dabcore_fops)) {
                err ("dabcore: unable to get major %d", DABCORE_MAJOR);
                return -EIO;
        }
	devfs_mk_dir("dabcore");
	
	dabcore_class = class_create(THIS_MODULE, "dabcore");
	
	if (IS_ERR(dabcore_class)) {
                err = PTR_ERR(dabcore_class);
                goto out_devfs;
        }

	for (i = 0; i < MAX_DABCORE_MINORS; i++) {
		dabcore_devs[i]=NULL;
		class_device_create(dabcore_class, NULL,
				    MKDEV(DABCORE_MAJOR, i),
				    NULL,
				    "dabcore%i",i);

		devfs_mk_cdev(MKDEV(DABCORE_MAJOR, i), S_IFCHR | S_IRUGO | S_IWUGO,
			      "dabcore/%d", i);
	}

	init_timer(&irq_emu_timer);

	if (emuirq) {
		irq_emu_timer.function=irq_emu_function;
		mod_timer(&irq_emu_timer, jiffies+ (HZ/20));
	}

        if (pci_module_init(&dabcore_driver)) {
		err=-EIO;
		goto out_class;
	}
        return 0;
out_class:
	class_destroy(dabcore_class);
out_devfs:
	devfs_remove("dabcore");
	unregister_chrdev(DABCORE_MAJOR,"dabcore");
        err("error register dabcore device");
        return err;
}

static void __exit dabcore_exit(void)
{
	int i;
        del_timer(&irq_emu_timer);
	pci_unregister_driver(&dabcore_driver); 
	unregister_chrdev(DABCORE_MAJOR, "dabcore");

	for (i = 0; i < MAX_DABCORE_MINORS; i++) {
                devfs_remove("dabcore/%d", i);
		class_device_destroy(dabcore_class, MKDEV(DABCORE_MAJOR, i));
	}

	devfs_remove("dabcore");
        class_destroy(dabcore_class);

        return;
}

module_init(dabcore_init);
module_exit(dabcore_exit);

MODULE_AUTHOR ("Deti Fliegl, Georg Acher, {fliegl|acher}@baycom.de");
MODULE_DESCRIPTION ("Dabcore Driver for 2.6");
module_param(buffers,int,32);
MODULE_PARM_DESC(buffers,"sets size n*4096 of internal buffers");
module_param(pciburstlen,int,1);
MODULE_PARM_DESC(pciburstlen,"sets pci burstlength (must be 1 < length < 25)");
module_param(emuirq,int,0);
MODULE_PARM_DESC(emuirq,"enables IRQ emulation in case of problems with shared PCI-IRQs");

MODULE_LICENSE("GPL");
